﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace EnumeratorStructures {

	class A<T> : IEnumerable<T> {
		public Enumerator GetEnumerator() {
			Console.WriteLine("+++ 111 +++ public A<T>.GetEnumerator()");
			return new Enumerator(42);
		}

		IEnumerator<T> IEnumerable<T>.GetEnumerator() {
			Console.WriteLine("+++ 222 +++ A<T>.IEnumerable<T>.GetEnumerator()");
			return GetEnumerator();
		}

		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator() {
			Console.WriteLine("+++ 333 +++ A<T>.IEnumerable.GetEnumerator()");
			return GetEnumerator();
		}

		public struct Enumerator : IEnumerator<T> {
			private int i;

			public Enumerator(int dummy) {
				i = 0;
			}

			public T Current {
				get {
					Console.WriteLine("--- A<T>.Enumerator.Current");
					return default(T);
				}
			}

			public void Dispose() {
			}

			object System.Collections.IEnumerator.Current {
				get { return Current; }
			}

			public bool MoveNext() {
				Console.Write("--- A<T>.Enumerator.MoveNext -> ");
				if (i > 1) {
					Console.WriteLine("false");
					return false;
				}

				i++;
				Console.WriteLine("true");
				return true;
			}

			public void Reset() {
				i = 0;
			}
		}
	}

	class Program {
		static void PrintAll(IEnumerable<int> e) {
			foreach (int i in e) {
				Console.WriteLine("Current == {0}", i);
			}
		}

		static void PrintGeneric<T>(T e) where T : IEnumerable<int> {
			foreach (int i in e) {
				Console.WriteLine("Current == {0}", i);
			}
		}

		static void Main(string[] args) {
			A<int> a = new A<int>();

			Console.WriteLine("Explicit a.GetEnumerator():");
			var e = a.GetEnumerator();
			e.MoveNext();
			Console.WriteLine("Current == {0}", e.Current);
			Console.WriteLine();

			/////
			
			Console.WriteLine("foreach over a:");
			foreach (int i in a) {
				Console.WriteLine("Current == {0}", i);
			}
			Console.WriteLine();
			
			/////
			
			Console.WriteLine("foreach over (IEnumerable<T>) a (via method PrintAll):");
			PrintAll(a);
			Console.WriteLine();

			/////

			Console.WriteLine("foreach over (IEnumerable<T>) a (via generic method PrintGeneric):");
			PrintGeneric(a);
			Console.WriteLine();
			
			/////
			
			// Testing an extension method (extension on IEnumerable<T>)
			Console.WriteLine("a.Count():");
			int count = a.Count();
			Console.WriteLine("Count == {0}", count);
			Console.WriteLine();
		}
	}
}
